/* =============================================================================
 UPC - EETAC - CSD - http://digsys.upc.edu 
  
 * Integrating MPLAB X + XC8 + MCC + Proteus VSM + Explorer 8 board
 * Target microcontroller: PIC18F46k22 
        https://www.microchip.com/en-us/product/PIC18F46K22 
 
 * Explorer 8 board user guide: 
        https://ww1.microchip.com/downloads/en/DeviceDoc/40001812B.pdf   

 * Materials adapted from LAB10 design phase #1: counter_BCD_1digit at: 
        https://digsys.upc.edu/csd/P10/lab10/lab10.html#exp8
        
 * Design phase #2: Adding the LCD to display ASCII text.Variable "Var_LCD_flag"
   will be used to print only when new information. 

 * Print in colour this file, and analyse it and use it as 'theory' class notes. 
  
============================================================================= */

/* 	============================================================================
    Board button S1 is INT0 and it is the external counter's CLK.  
    S2 button is CE_L control input.
    LEDs RB3 RB2

  * Explorer 8 LCD display is installed by default through an MCP23S17
 *  16-Bit I/O Expander with Serial SPI Interface 
   	========================================================================= */

/* 	============================================================================
 IMPORTANT NOTE: This code is used with MCC (MPLAB code configurator) 
 * Do not update or force update main.c and main.h files because 
 * Be aware of what to update for other c and h files generated by MCC
   	========================================================================= */

// MCC generated application headers 
#include "mcc_generated_files/mcc.h"
#include "lcd.h"
// Application header including global variables and function prototypes
#include "main.h" //

/* 	============================================================================
	Main function 
  	========================================================================= */
void main(void){
    init_system ();
    while(1) {                      /* loop forever the FSM */
		if (state_logic()) break;
		if (output_logic()) break;
    }
    while (1){
        __nop();//The system has crashed in an illegal state and needs attention
                //This second infinite loop is never reached in normal operation
    }           // The only way to scape is clicking reset (CD)
}	


/* 	============================================================================
	Function definitions
  	========================================================================= */

//*****************************************************************************
//State variable logic routine  (The CC1)
//*****************************************************************************
char state_logic(void){
	char error = 0;	

    read_inputs(); //sample input values so that the CC1 can decide transitions
    
	if (Var_CLK_flag == 1){
    	switch (current_state) {
            case Num_0:				
                if (Var_CE == 1){
                    current_state = Num_1;     
                    Var_LCD_flag = 1;  // Print a new message only when changing                         
                }
                else{
                    current_state = Num_0;  
                }                    
            break;

            case Num_1:				
                if (Var_CE == 1){ 
                    current_state = Num_2;     
                    Var_LCD_flag = 1;  // Print a new message only when changing                                  
                }
                else current_state = Num_1;                    
            break;
            
            case Num_2:		//else is not required. 
                            // Check if  disassembly listing is shorter
                if (Var_CE == 1){ 
                    current_state = Num_3;   
                    Var_LCD_flag = 1;  // Print a new message only when changing                                  
                }
            break;

            case Num_3:				
                if (Var_CE == 1){ 
                    current_state = Num_4; 
                    Var_LCD_flag = 1;  // Print a new message only when changing                                  
                }
            break;
            case Num_4:				
                if (Var_CE == 1) {
                    current_state = Num_5;   
                    Var_LCD_flag = 1;  // Print a new message only when changing                                  
                }
           
            break;
            case Num_5:				
                if (Var_CE == 1) {
                    current_state = Num_6;     
                    Var_LCD_flag = 1;  // Print a new message only when changing                                  
                }
            break;
            case Num_6:				
                if (Var_CE == 1) {
                    current_state = Num_7;  
                    Var_LCD_flag = 1;  // Print a new message only when changing                                  
                }
            break;
            case Num_7:				
                if (Var_CE == 1) {
                    current_state = Num_8;    
                    Var_LCD_flag = 1;  // Print a new message only when changing                                  
                }
            break;
            case Num_8:				
                if (Var_CE == 1) {
                    current_state = Num_9;    
                    Var_LCD_flag = 1;  // Print a new message only when changing                                  
                }
            break;
            case Num_9:				
                if (Var_CE == 1) {
                    current_state = Num_0;                    
                    Var_LCD_flag = 1;  // Print a new message only when changing                                  
                }
            break;

            default:
            // If no states have been selected, there is an error
            // and the systems must be stopped                     
                error = 1;	
                break;
        }
		Var_CLK_flag = 0;	// Clear the Var_CLK_Flag detection flag
	}
	return (error);
}
//*****************************************************************************
//Output logic routine (The combinational circuit CC2)
//*****************************************************************************
char output_logic(void){
unsigned char error = 0;	 
	switch (current_state) {	
		case Num_0:
            Var_BCD_code = 0;
            
            if (Var_LCD_flag == 1) {//write only if there is a new info to print
					// Write the output pins once the output variables are set
                LCD_GoTo(0,0); // first row
                LCD_WriteString(" BCD up counter.");                     
                LCD_GoTo(1,0); // second row
                LCD_WriteString("... Number 0 ...");                
                Var_LCD_flag = 0;  // Clear flag, so that no new rewriting                 
            }
			break;
		case Num_1:
            Var_BCD_code = 1;
            if (Var_LCD_flag == 1) {
                LCD_GoTo(1,0); // second row
                LCD_WriteString("... Number 1 ...");                
                Var_LCD_flag = 0;  
            }
			break;
		case Num_2:
            Var_BCD_code = 2;
            if (Var_LCD_flag == 1) {
                LCD_GoTo(1,0); 
                LCD_WriteString("... Number 2 ...");                
                Var_LCD_flag = 0;   
            }
			break;           
			break;
		case Num_3:
            Var_BCD_code = 3;
            if (Var_LCD_flag == 1) {
                LCD_GoTo(1,0); 
                LCD_WriteString("... Number 3 ...");                
                Var_LCD_flag = 0;  
            }            
			break;
		case Num_4:
            Var_BCD_code = 4;
            if (Var_LCD_flag == 1) {
                LCD_GoTo(1,0); 
                LCD_WriteString("... Number 4 ...");                
                Var_LCD_flag = 0;   
            }              
			break;
		case Num_5:
            Var_BCD_code = 5;
            if (Var_LCD_flag == 1) {
                LCD_GoTo(1,0); 
                LCD_WriteString("... Number 5 ...");                
                Var_LCD_flag = 0;  
            }              
			break;
		case Num_6:
            Var_BCD_code = 6;
            if (Var_LCD_flag == 1) {
                LCD_GoTo(1,0);
                LCD_WriteString("... Number 6 ...");                
                Var_LCD_flag = 0;   
            }              
			break;		case Num_7:
            Var_BCD_code = 7;
            if (Var_LCD_flag == 1) {
                LCD_GoTo(1,0); // second row
                LCD_WriteString("... Number 7 ...");                
                Var_LCD_flag = 0;   
            }               
			break;
		case Num_8:
            Var_BCD_code = 8;
            if (Var_LCD_flag == 1) {
                LCD_GoTo(1,0); // second row
                LCD_WriteString("... Number 8 ...");                
                Var_LCD_flag = 0;    
            }               
			break;
		case Num_9:
            Var_BCD_code = 9;
            if (Var_LCD_flag == 1) {
                LCD_GoTo(1,0); // second row
                LCD_WriteString("... Number 9 ...");                
                Var_LCD_flag = 0;  
            }               
			break;
		default:
			error = 1;	// If no states have been selected, there is an error
						// and the systems must be stopped 
			break;
	}
// Notice that '&' is a bitwise AND operation, however '&&' is a logic AND: 
	if (current_state == Num_9 && Var_CE == 1)
		Var_TC10 = 1;
	else
		Var_TC10 = 0;

    write_outputs(); // The interface with the hardware to convert variables 
					 // into voltages at a given microcontroller pin. 
    
	return (error);

}
/* =============================================================================
 Scan inputs and process the bits so that we can set our convenient variables 
 many of the variables are "char" because we can easily monitor them using
 the watch window for debugging purposes.
  =========================================================================== */
void read_inputs (void){     
// Like in P9, read the port, mask the bit of interest and set the variable 
    uint8_t var_buf; 
    var_buf = (PORTA & 0b00100000)>>5; //Active_low value read
    Var_CE = (!var_buf)& 0b00000001; // bitwise not
    // S2 connected at PORTA bit 5 is CE_L
}

/* =============================================================================
	Driving the output pins. Let's drive the output ports, converting 
	microcontroller variables into electrical signals. 
 * In Explorer 8 we have 8 LED.
 * PORTD(3..0) will be Q(3..0) the counter's BCD output code   
 * PORTB3 will be the counter's terminal count TC10
   ========================================================================== */
void write_outputs(void){
	uint8_t Var_buf; 
    //To write a given output pin:
    // Read and mask to preserve all the bits not to be written
    Var_buf = PORTB & 0b11110111; 
    // Shift the bit of interest to the pin position, compose the port bits (OR)
    // and write the port
    PORTB = Var_buf | (Var_TC10 << 3);  // TC10 is PORTB bit 3

    //The same procedure to write the Q(3..0) to the assigned PORTD pins (3..0)
    Var_buf = PORTD & 0b11110000; 
    PORTD = Var_buf | Var_BCD_code;
}


/* =============================================================================
	Init system: Initialise the microcontroller system 
   ========================================================================== */
void init_system (void){
    
    SYSTEM_Initialize();  //Configuration from MCC
    LCD_Initialize();     // Default LCD on Explorer 8
    
    // Enable global interrupts to use INT0
    INTERRUPT_GlobalInterruptEnable(); 
    
    // Initial message to check that hardware and software is ok
    LCD_GoTo(0,0);
    LCD_WriteString("  CSD - P11     ");
    LCD_GoTo(1,0);
    LCD_WriteString("  BCD counter   ");
    
}